import asyncio
import logging
import socket


logger = logging.getLogger(__name__)


class TcpServer(object):
    """A simple asyncio TCP server

    使用 loop 的 sock_accept/sock_sendall/sock_recv 等方法来实现，
    这些方法相对都比较 low-level。

    asyncio 也提供了 ``loop.create_server`` 方法创建一个 server，然后使用
    StreamReader 和 StreamWriter 来进行读写，整体来说更加 high level。
    StreamReader 封装了类似 readline/readuntil 等方法，使用起来很方便。
    StreamWriter 有一个自己的 buffer，所以调用 write 方法时，
    它没有真正的将数据写给操作系统，在 3.7 之前，它似乎都没有很好的办法来确保这一点。
    不过它有个 drain 方法来将 buffer 中的一部分数据 flush 到操作系统。

    总体来说，个人目前更喜欢这个 TcpServer 的实现，接口非常明确，
    和阻塞的 socket 使用起来特别像。而 asyncio 提供的那一套用起来不直观，
    不过现在 feeluown 中使用的正是 asyncio 这一套，大概是强迫症。
    """

    def __init__(self, host, port, handle_func):
        self.host = host
        self.port = port
        self.handle_func = handle_func

        self._sock = None

    async def run(self, *args, **kwargs):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.host, self.port))
        sock.listen()
        # make restart easier
        sock.setblocking(0)
        self._sock = sock

        event_loop = asyncio.get_event_loop()
        while True:
            conn, addr = await event_loop.sock_accept(sock)
            event_loop.create_task(
                self.handle_func(conn, addr, *args, **kwargs))



from feeluown.app.gui_app import GuiApp
from feeluown.collection import CollectionType, Collection
from feeluown.library import ModelType
from feeluown.utils.reader import wrap
from feeluown.gui.page_containers.table import Renderer

from feeluown.gui.base_renderer import TabBarRendererMixin


async def render(req, identifier, **kwargs):
    app: GuiApp = req.ctx["app"]
    ui = app.ui
    tab_index = int(req.query.get("tab_index", 0))

    coll = app.coll_mgr.get(identifier)

    mixed = False
    model_type = None
    if coll.type is CollectionType.sys_library:
        mixed = True
    else:
        types = set()
        for model in coll.models:
            types.add(model.meta.model_type)
            if len(types) >= 2:
                mixed = True
                break
        if types:
            model_type = types.pop()
        else:
            model_type = ModelType.song

    if mixed is True:
        ui.right_panel.set_body(ui.right_panel.scrollarea)
        table_container = ui.right_panel.table_container
        renderer = LibraryRenderer(app, tab_index, coll)
        await table_container.set_renderer(renderer)
    else:
        ui.right_panel.show_collection(coll, model_type)


class LibraryRenderer(Renderer, TabBarRendererMixin):
    """
    This Renderer was used to render library collection particularlly.

    TODO: create a more elegant renderer for mixed collection.
    NOTE(cosven): I think mixed collection should be rendered in single page
    without tab.
    """

    def __init__(self, app, tab_index, coll: Collection):
        self._app = app
        self._coll = coll
        self.tab_index = tab_index
        self.tabs = self.default_tabs()

    async def render(self):
        coll = self._coll
        self.meta_widget.show()
        if coll.type is CollectionType.sys_library:
            self.meta_widget.title = "音乐库"
        else:
            self.meta_widget.title = coll.name

        if not coll.models:
            # HACK: show a message by meta widget
            if coll.type is CollectionType.sys_library:
                self.meta_widget.source = "音乐库暂无收藏，快去添加一些吧！"
            else:
                self.meta_widget.source = "该收藏集暂无内容，快去添加一些吧！"
            return

        self.render_tab_bar()
        self.render_models()

        def remove_song(model):
            # TODO(cosven): the whole UI is refreshed after a model is removed,
            # ideally, only part of the UI is refreshed. For example, a user scroll
            # to the bottom of the list view, when the last model is removed,
            # the UI is refreshed and the the user needs to scroll to the bottom again.
            self._coll.remove(model)
            self.render_models()
            self._app.show_msg(f"移除 {model} 成功")

        def remove_model(model, cb):
            self._coll.remove(model)
            cb(model, True)
            self._app.show_msg(f"移除 {model} 成功")

        if self.tabs[self.tab_index][1] is ModelType.song:
            self.songs_table.remove_song_func = remove_song
        self.artists_table.enable_remove_action = True
        self.albums_table.enable_remove_action = True
        self.playlists_table.enable_remove_action = True
        self.videos_table.enable_remove_action = True
        self.artists_table.remove_item_needed.connect(remove_model)
        self.albums_table.remove_item_needed.connect(remove_model)
        self.playlists_table.remove_item_needed.connect(remove_model)
        self.videos_table.remove_item_needed.connect(remove_model)

    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(
            page=f"/colls/{self._coll.identifier}", query={"tab_index": tab_index}
        )

    def render_models(self):
        _, mtype, show_handler = self.tabs[self.tab_index]
        models = [
            model for model in self._coll.models if model.meta.model_type == mtype
        ]
        reader = wrap(models)
        show_handler(reader)



from PyQt6.QtCore import QTimer, QSize
from PyQt6.QtGui import QColor, QPalette
from PyQt6.QtWidgets import QVBoxLayout, QPushButton, QWidget, QHBoxLayout, QLabel
from PyQt6.QtOpenGLWidgets import QOpenGLWidget

from feeluown.player import State
from feeluown.gui.widgets import PlayPauseButton
from feeluown.gui.widgets.progress_slider import ProgressSlider
from feeluown.gui.widgets.size_grip import SizeGrip
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.labels import ProgressLabel, DurationLabel


class Button(TextButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        font = self.font()
        font.setPixelSize(10)
        self.setFont(font)
        self.setFixedHeight(16)

    def sizeHint(self):
        fm = self.fontMetrics()
        rect = fm.boundingRect(self.text())
        return QSize(rect.width() + 6, rect.height())


class VideoPlayerCtlBar(QWidget):
    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        self._app = app

        # Create widgets.
        self._toggle_btn = PlayPauseButton(draw_circle=False)
        self._progress_slider = ProgressSlider(app)
        self._progress_label = ProgressLabel(app, self)
        self._duration_label = DurationLabel(app, self)
        self._size_grip = SizeGrip(parent=self)
        self._adhoc_btn_layout = QHBoxLayout()
        self._layout = QVBoxLayout(self)
        self._bottom_layout = QHBoxLayout()

        # Do initialization for widgets.
        self._toggle_btn.setCheckable(True)
        self._setup_ui()

        self._app.player.state_changed.connect(
            lambda x: self._toggle_btn.setChecked(x == State.playing),
            weak=False,
            aioqueue=True,
        )
        self._toggle_btn.clicked.connect(self._app.player.toggle)

    def add_adhoc_btn(self, text: str) -> QPushButton:
        """
        .. versionadded: 4.0.1
        """
        btn = Button(text)
        self._adhoc_btn_layout.addWidget(btn)
        btn.setPalette(self.palette())
        return btn

    def clear_adhoc_btns(self):
        """
        .. versionadded: 4.0.1
        """
        while self._adhoc_btn_layout.count():
            item = self._adhoc_btn_layout.itemAt(0)
            self._adhoc_btn_layout.removeItem(item)
            if item.widget():
                item.widget().deleteLater()

    def _setup_ui(self):
        self.setAutoFillBackground(True)

        # Setup layout.
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._progress_slider)
        self._layout.addLayout(self._bottom_layout)

        # Setup bottom layout.
        self._layout.addStretch(0)
        self._bottom_layout.setSpacing(2)
        self._adhoc_btn_layout.setSpacing(4)
        self._bottom_layout.addWidget(self._toggle_btn)
        self._bottom_layout.addSpacing(6)
        self._bottom_layout.addWidget(self._progress_label)
        self._bottom_layout.addWidget(QLabel("/", self))
        self._bottom_layout.addWidget(self._duration_label)
        self._bottom_layout.addStretch(1)
        self._bottom_layout.addLayout(self._adhoc_btn_layout)
        self._bottom_layout.addSpacing(6)
        self._bottom_layout.addWidget(self._size_grip)

        # Setup widgets size.
        self._size_grip.setFixedSize(20, 20)

        # Customize the palette.
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Text, QColor("white"))
        palette.setColor(QPalette.ColorRole.ButtonText, QColor("white"))
        palette.setColor(QPalette.ColorRole.WindowText, QColor("white"))
        bg_color = QColor("black")
        bg_color.setAlpha(180)  # Make it semi-transparent.
        palette.setColor(QPalette.ColorRole.Window, bg_color)
        self.setPalette(palette)
        # It seems children don't inherit the palette, so we set palette for
        # children explicitly.
        for child in self.findChildren(QWidget):
            child.setPalette(palette)


class VideoOpenGLWidget(QOpenGLWidget):
    """Base class for widget which provides OpenGL based video rendering

    This widget has a overlay which currently contains a player control bar.
    """

    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        # Define public attributes.
        self.overlay_auto_visible = True

        # Define protected attributes.
        self._gl_painters = set({})
        self._overlay_visible_timer = QTimer(self)
        self._overlay_visible_duration = 2000  # 2s

        # Define ui related objects and attributes.
        self.ctl_bar = VideoPlayerCtlBar(app=app, parent=self)
        self._layout = QVBoxLayout(self)

        # Do initialization.
        self.setMouseTracking(True)
        self._overlay_visible_timer.timeout.connect(self.hide_overlay)
        self._setup_ui()

    def _setup_ui(self):
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addStretch(1)
        self._layout.addWidget(self.ctl_bar)
        self.hide_overlay()

    def add_gl_painter(self, painter):
        """Add QPainter based gl painter.

        In paintGL, we invoke these painters one by one so that they can
        paint their own contents.

        :param painter: painter must impl `paint(opengl_widget)` method.

        .. versionadded: 3.7.13
        """
        # Simple validation
        assert hasattr(painter, "paint"), "painter must impl `paint` method."
        self._gl_painters.add(painter)

    def hide_overlay(self):
        """hide the overlay widget"""
        self.ctl_bar.hide()

    def show_overlay(self):
        """show the overlay widget"""
        self.ctl_bar.show()

    def mouseMoveEvent(self, e):
        super().mouseMoveEvent(e)
        if self.overlay_auto_visible is True:
            self.show_overlay()
            self._overlay_visible_timer.start(self._overlay_visible_duration)

    def enterEvent(self, e):
        super().enterEvent(e)
        if self.overlay_auto_visible is True:
            self.show_overlay()
            self._overlay_visible_timer.start(self._overlay_visible_duration)

    def leaveEvent(self, e):
        super().leaveEvent(e)
        self._overlay_visible_timer.start(self._overlay_visible_duration)


if __name__ == "__main__":
    from feeluown.gui.debug import simple_layout, mock_app

    with simple_layout(theme="dark") as layout, mock_app() as app:
        ctl_bar = VideoPlayerCtlBar(app)
        ctl_bar.add_adhoc_btn("全屏")
        ctl_bar.add_adhoc_btn("最小化")
        layout.addWidget(ctl_bar)



"""
encapsulate asyncio API
~~~~~~~~~~~~~~~~~~~~~~~

Asyncio has added some new API in Python 3.7 and 3.8, e.g.,
gather, run and create_task. They are recommended in newer version.
However, FeelUOwn is supposed to be compatible with Python 3.5,
we will backport some API here. In addition, we create alias
for some frequently used API.
"""

import asyncio
import sys


if sys.version_info >= (3, 7):
    # create_task is temporarily not working properly with quamash
    # https://github.com/harvimt/quamash/issues/116
    create_task = asyncio.ensure_future
else:
    create_task = asyncio.ensure_future

#: alias for create_task
spawn = create_task
as_completed = asyncio.as_completed

#: sleep is an alias of `asyncio.sleep`.
sleep = asyncio.sleep

#: run is an alias of `asyncio.run`.
run = asyncio.run

#: wait is an alias of `asyncio.wait`.
wait = asyncio.wait

#: gather is an alias of `asyncio.gather`.
gather = asyncio.gather

#: wait_for is an alias of `asyncio.wait_for`
wait_for = asyncio.wait_for

bg_tasks = set()


def run_in_executor(executor, func, *args):
    """alias for loop.run_in_executor"""
    loop = asyncio.get_running_loop()
    return loop.run_in_executor(executor, func, *args)


def run_afn(afn, *args):
    """Alias for create_task

    .. versionadded:: 3.7.8
    """
    return create_task(afn(*args))


def run_afn_ref(afn, *args):
    """Create a background task and keep a reference.

    .. versionadded:: 4.1.9
    """
    task = create_task(afn(*args))
    bg_tasks.add(task)
    task.add_done_callback(bg_tasks.discard)
    return task


def run_fn(fn, *args):
    """Alias for run_in_executor with default executor

    .. versionadded:: 3.7.8
    """
    return run_in_executor(None, fn, *args)



import logging
from typing import TYPE_CHECKING, Optional

from PyQt6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout
from feeluown.library.provider_protocol import SupportsToplist

from feeluown.utils.reader import create_reader
from feeluown.utils.aio import run_fn
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.img_card_list import (
    PlaylistCardListView,
    PlaylistCardListModel,
    PlaylistFilterProxyModel,
    PlaylistCardListDelegate,
)

from feeluown.library import (
    SupportsRecListDailyPlaylists,
    SupportsRecListDailySongs,
    SupportsCurrentUserDislikeSongsReader,
    SupportsCurrentUserListRadioSongs,
)

from feeluown.gui.widgets import CalendarButton, RankButton, EmojiButton


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


logger = logging.getLogger(__name__)


async def render(req, **kwargs):
    app: "GuiApp" = req.ctx["app"]

    view = View(app)
    app.ui.right_panel.set_body(view)
    await view.render()


class View(QWidget):
    def __init__(self, app: "GuiApp"):
        super().__init__(parent=None)
        self._app = app
        self._heart_radar_provider: Optional[SupportsCurrentUserListRadioSongs] = None

        self.header_title = LargeHeader()
        self.header_playlist_list = MidHeader()
        self.playlist_list_view = PlaylistCardListView(fixed_row_count=1)
        self.playlist_list_view.setItemDelegate(
            PlaylistCardListDelegate(
                self.playlist_list_view,
                card_min_width=100,
            )
        )
        self.daily_songs_btn = CalendarButton("每日推荐", parent=self)
        self.rank_btn = RankButton(parent=self)
        # FIXME: design a new button for dislike
        self.dislike_btn = EmojiButton("🚫", "音乐黑名单", parent=self)
        self.heart_radar_btn = EmojiButton("📻", "红心雷达", parent=self)
        self.daily_songs_btn.setMinimumWidth(150)
        self.rank_btn.setMinimumWidth(150)
        self.dislike_btn.setMinimumWidth(150)
        self.heart_radar_btn.setMinimumWidth(150)

        self.header_title.setText("发现音乐")
        self.header_playlist_list.setText("个性化推荐")

        self._layout = QVBoxLayout(self)
        self._setup_ui()

        self.playlist_list_view.show_playlist_needed.connect(
            lambda model: self._app.browser.goto(model=model)
        )
        self.daily_songs_btn.clicked.connect(
            lambda: self._app.browser.goto(page="/rec/daily_songs")
        )
        self.rank_btn.clicked.connect(lambda: self._app.browser.goto(page="/toplist"))
        self.heart_radar_btn.clicked.connect(self._on_heart_radar_clicked)
        self.dislike_btn.clicked.connect(
            lambda: self._app.browser.goto(page="/my_dislike")
        )

    def _setup_ui(self):
        self._h_layout = QHBoxLayout()
        self._h_layout.addWidget(self.daily_songs_btn)
        self._h_layout.addSpacing(10)
        self._h_layout.addWidget(self.rank_btn)
        self._h_layout.addSpacing(10)
        self._h_layout.addWidget(self.heart_radar_btn)
        self._h_layout.addSpacing(10)
        self._h_layout.addWidget(self.dislike_btn)
        self._h_layout.addStretch(0)

        self._layout.setContentsMargins(20, 10, 20, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.header_title)
        self._layout.addSpacing(10)
        self._layout.addLayout(self._h_layout)
        self._layout.addSpacing(30)
        self._layout.addWidget(self.header_playlist_list)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.playlist_list_view)
        self._layout.addStretch(0)

    async def render(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        if pvd_ui is None:
            self._app.show_msg("wtf!")
            return

        provider = pvd_ui.provider
        if isinstance(provider, SupportsRecListDailyPlaylists):
            playlists = await run_fn(provider.rec_list_daily_playlists)
            model = PlaylistCardListModel.create(create_reader(playlists), self._app)
            filter_model = PlaylistFilterProxyModel()
            filter_model.setSourceModel(model)
            self.playlist_list_view.setModel(filter_model)

        self.daily_songs_btn.setEnabled(isinstance(provider, SupportsRecListDailySongs))
        self.rank_btn.setEnabled(isinstance(provider, SupportsToplist))
        self.heart_radar_btn.setEnabled(
            isinstance(provider, SupportsCurrentUserListRadioSongs)
        )
        self.dislike_btn.setEnabled(
            isinstance(provider, SupportsCurrentUserDislikeSongsReader)
        )

    def _on_heart_radar_clicked(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        assert pvd_ui is not None
        provider = pvd_ui.provider
        assert isinstance(provider, SupportsCurrentUserListRadioSongs)

        was_active = self._app.fm.is_active
        current_provider = self._heart_radar_provider if was_active else None
        fetch_func = provider.current_user_list_radio_songs
        if was_active and current_provider is provider:
            self._app.show_msg("红心雷达已激活")
            return

        if was_active:
            self._app.fm.deactivate()

        self._app.fm.activate(fetch_func)
        self._heart_radar_provider = provider
        self._app.show_msg("红心雷达已切换" if was_active else "红心雷达已激活")



from PyQt6.QtCore import Qt, QRect, QEvent, QPointF
from PyQt6.QtWidgets import (
    QWidget,
    QStackedLayout,
    QVBoxLayout,
    QHBoxLayout,
    QApplication,
)
from PyQt6.QtGui import (
    QColor,
    QLinearGradient,
    QPalette,
    QPainter,
)

from feeluown.player import PlaybackMode, SongsRadio
from feeluown.gui.helpers import fetch_cover_wrapper, esc_hide_widget
from feeluown.gui.components.player_playlist import PlayerPlaylistView
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.tabbar import TabBar
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListView,
    SongMiniCardListModel,
    SongMiniCardListDelegate,
)
from feeluown.utils.reader import create_reader


PlaybackModeName = {
    PlaybackMode.one_loop: "单曲循环",
    PlaybackMode.sequential: "顺序播放",
    PlaybackMode.loop: "循环播放",
    PlaybackMode.random: "随机播放",
}
PlaybackModes = list(PlaybackModeName.keys())


def acolor(s, a):
    """Create color with it's name and alpha"""
    color = QColor(s)
    color.setAlpha(a)
    return color


class PlaylistOverlay(QWidget):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._app = app
        self._tabbar = TabBar(self)
        self._clear_playlist_btn = TextButton("清空播放队列")
        self._playback_mode_switch = PlaybackModeSwitch(app)
        self._goto_current_song_btn = TextButton("跳转到当前歌曲")
        self._songs_radio_btn = TextButton("自动续歌")
        # Please update the list when you add new buttons.
        self._btns = [
            self._clear_playlist_btn,
            self._playback_mode_switch,
            self._goto_current_song_btn,
            self._songs_radio_btn,
        ]
        self._stacked_layout = QStackedLayout()
        self._shadow_width = 15
        self._view_options = dict(row_height=60, no_scroll_v=False)
        self._player_playlist_view = PlayerPlaylistView(self._app, **self._view_options)

        # AutoFillBackground should be disabled for PlaylistOverlay so that shadow
        # effects can be simulated. AutoFillBackground should be enabled for tabbar.
        self._tabbar.setAutoFillBackground(True)

        self._clear_playlist_btn.clicked.connect(self._app.playlist.clear)
        self._goto_current_song_btn.clicked.connect(self.goto_current_song)
        self._songs_radio_btn.clicked.connect(self.enter_songs_radio)
        esc_hide_widget(self)
        q_app = QApplication.instance()
        assert q_app is not None  # make type checker happy.
        # type ignore: q_app has focusChanged signal, but type checker can't find it.
        q_app.focusChanged.connect(self.on_focus_changed)  # type: ignore
        self._app.installEventFilter(self)
        self._tabbar.currentChanged.connect(self.show_tab)

        self.setup_ui()

    def setup_ui(self):
        self._layout = QVBoxLayout(self)
        self._btn_layout = QHBoxLayout()
        self._btn_layout2 = QHBoxLayout()
        self._layout.setContentsMargins(self._shadow_width, 0, 0, 0)
        self._layout.setSpacing(0)
        self._btn_layout.setContentsMargins(7, 7, 7, 7)
        self._btn_layout.setSpacing(7)
        self._btn_layout2.setContentsMargins(7, 0, 7, 7)
        self._btn_layout2.setSpacing(7)

        self._tabbar.setDocumentMode(True)
        self._tabbar.addTab("播放列表")
        self._tabbar.addTab("最近播放")
        self._layout.addWidget(self._tabbar)
        self._layout.addLayout(self._btn_layout)
        self._layout.addLayout(self._btn_layout2)
        self._layout.addLayout(self._stacked_layout)

        self._btn_layout.addWidget(self._clear_playlist_btn)
        self._btn_layout.addWidget(self._playback_mode_switch)
        self._btn_layout.addWidget(self._goto_current_song_btn)
        self._btn_layout2.addWidget(self._songs_radio_btn)
        self._btn_layout.addStretch(0)
        self._btn_layout2.addStretch(0)

    def on_focus_changed(self, _, new):
        """
        Hide the widget when it loses focus.
        """
        if not self.isVisible():
            return
        # When the app is losing focus, the new is None.
        if new is None or new is self or new in self.findChildren(QWidget):
            return
        self.hide()

    def goto_current_song(self):
        view = self._stacked_layout.currentWidget()
        assert isinstance(view, PlayerPlaylistView)
        view.scroll_to_current_song()

    def enter_songs_radio(self):
        songs = self._app.playlist.list()
        if not songs:
            self._app.show_msg("播放队列为空，不能激活“自动续歌”功能")
        else:
            radio = SongsRadio(self._app, songs)
            self._app.fm.activate(radio.fetch_songs_func, reset=False)
            self._app.show_msg("“自动续歌”功能已激活")

    def show_tab(self, index):
        if not self.isVisible():
            return

        if index == 0:
            self._show_btns()
            view = self._player_playlist_view
        else:
            self._hide_btns()
            model = SongMiniCardListModel(
                create_reader(self._app.recently_played.list_songs()),
                fetch_cover_wrapper(self._app),
            )
            view = SongMiniCardListView(**self._view_options)
            view.setModel(model)
            view.play_song_needed.connect(self._app.playlist.play_model)
        delegate = SongMiniCardListDelegate(
            view,
            card_min_width=self.width() - self.width() // 6,
            card_height=40,
            card_padding=(5 + SongMiniCardListDelegate.img_padding, 5, 0, 5),
            card_right_spacing=10,
        )
        view.setItemDelegate(delegate)
        self._stacked_layout.addWidget(view)
        self._stacked_layout.setCurrentWidget(view)

    def _hide_btns(self):
        for btn in self._btns:
            btn.hide()

    def _show_btns(self):
        for btn in self._btns:
            btn.show()

    def paintEvent(self, e):
        super().paintEvent(e)

        painter = QPainter(self)
        painter.setPen(Qt.PenStyle.NoPen)

        # Draw shadow effect on the left side.
        painter.save()
        shadow_width = self._shadow_width
        rect = QRect(0, 0, shadow_width, self.height())
        gradient = QLinearGradient(QPointF(rect.topRight()), QPointF(rect.topLeft()))
        gradient.setColorAt(0, acolor("black", 70))
        gradient.setColorAt(0.05, acolor("black", 60))
        gradient.setColorAt(0.1, acolor("black", 30))
        gradient.setColorAt(0.2, acolor("black", 5))
        gradient.setColorAt(1, acolor("black", 0))
        painter.setBrush(gradient)
        painter.drawRect(rect)
        painter.restore()

        # Draw a rect to fill the remain background.
        painter.setBrush(self.palette().color(QPalette.ColorRole.Base))
        painter.drawRect(shadow_width, 0, self.width() - shadow_width, self.height())

    def showEvent(self, e):
        super().showEvent(e)
        self.show_tab(self._tabbar.currentIndex())

    def eventFilter(self, obj, event):
        """
        Hide myself when the app is resized.
        """
        if obj is self._app and event.type() == QEvent.Type.Resize:
            self.hide()
        return False


class PlaybackModeSwitch(TextButton):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

        self.update_text()
        self.clicked.connect(self.switch_playback_mode)
        self._app.playlist.playback_mode_changed.connect(
            self.on_playback_mode_changed, aioqueue=True
        )
        self.setToolTip("修改播放模式")

    def switch_playback_mode(self):
        playlist = self._app.playlist
        index = PlaybackModes.index(playlist.playback_mode)
        if index < len(PlaybackModes) - 1:
            new_index = index + 1
        else:
            new_index = 0
        playlist.playback_mode = PlaybackModes[new_index]

    def update_text(self):
        self.setText(PlaybackModeName[self._app.playlist.playback_mode])

    def on_playback_mode_changed(self, _):
        self.update_text()



from feeluown.library import ModelType, SupportsPlaylistSongsReader
from feeluown.library import resolve, reverse
from .base import AbstractHandler


class PlaylistHandler(AbstractHandler):
    cmds = ('add', 'remove', 'list', 'next', 'previous', 'clear',)

    def handle(self, cmd):  # pylint: disable=inconsistent-return-statements
        # pylint: disable=no-else-return
        if cmd.action == 'add':
            return self.add(cmd.args)
        elif cmd.action == 'remove':
            return self.remove(cmd.args[0].strip())
        elif cmd.action == 'clear':
            return self.clear()
        elif cmd.action == 'list':
            return self.list()
        elif cmd.action == 'next':
            self.playlist.next()
        elif cmd.action == 'previous':
            self.playlist.previous()

    def add(self, furi_list):
        playlist = self.playlist
        for furi in furi_list:
            furi = furi.strip()
            obj = resolve(furi)
            if obj is not None:
                obj_type = obj.meta.model_type
                if obj_type == ModelType.song:
                    playlist.add(obj)
                elif obj_type == ModelType.playlist:
                    provider = self.library.get(obj.source)
                    if isinstance(provider, SupportsPlaylistSongsReader):
                        reader = provider.playlist_create_songs_rd(obj)
                        for song in reader:
                            playlist.add(song)
                    # TODO: raise error if it does not support

    def remove(self, song_uri):
        # FIXME: a little bit tricky
        for song in self.playlist.list():
            if reverse(song) == song_uri:
                self.playlist.remove(song)
                break

    def list(self):
        songs = self.playlist.list()
        return songs

    def clear(self):
        self.playlist.clear()



from PyQt6.QtCore import Qt, QRectF, QPoint
from PyQt6.QtGui import QTextOption
from PyQt6.QtWidgets import QToolTip

from feeluown.gui.widgets.statusline import StatuslineLabel


class NotifyStatus(StatuslineLabel):
    def __init__(self, app, parent=None):
        super().__init__(app, parent)
        self._app = app

    def drawInner(self, painter):
        inner_rect = QRectF(0, 0, self._inner_width, self._inner_height)
        painter.drawText(inner_rect, "✉", QTextOption(Qt.AlignmentFlag.AlignCenter))

    def show_msg(self, text, timeout=1500):
        QToolTip.showText(
            self.mapToGlobal(QPoint(0, 0)), text, self, self.rect(), timeout
        )
